﻿using UnityEngine;
using System;
using System.Collections.Generic;
using static MacGruber.Breathing;
using static MacGruber.Utils;


namespace MacGruber
{
	public static partial class Utils
    {
        static Utils()
        {
            PackagePathOverride = "MacGruber.Life.13:/";  // <= Make sure to refer to the correct package here!
            PluginPathOverride = PackagePathOverride + "Custom/Scripts/MacGruber/Life";
            IsInPackageOverride = true;
        }
    }
	
	public class DriverBreathing : MVRScript
	{
		private Breathing myBreathing;

		private JSONStorableFloat myStomachMin;
		private JSONStorableFloat myStomachMax;
		private JSONStorableFloat myChestJointMin;
		private JSONStorableFloat myChestJointMax;
		private JSONStorableFloat myChestMorphMin;
		private JSONStorableFloat myChestMorphMax;
		private JSONStorableFloat myChestSpring;
		private JSONStorableFloat myMouthOpenMin;
		private JSONStorableFloat myMouthOpenMax;
		private JSONStorableFloat myMouthOpenTime;
		private JSONStorableFloat myMouthCloseTime;
		private JSONStorableFloat myLipsMax;
		private JSONStorableFloat myNoseInMax;
		private JSONStorableFloat myNoseOutMax;
		private JSONStorableFloat myDebug1;
		private JSONStorableFloat myDebug2;
		private JSONStorableFloat myDebug3;

		DAZCharacterSelector myGeometry;
		DAZCharacterSelector.Gender myGender = DAZCharacterSelector.Gender.None;
		private DAZMorph myChestMorph;
		private DAZMorph myStomachMorph;
		private DAZMorph myMouthMorph;
		private DAZMorph myLipsMorph;
		private DAZMorph myNoseInMorph;
		private DAZMorph myNoseOutMorph;
		private FreeControllerV3 myChestControl;
		private bool myAnimationValid = false;
		private float myMouthOpen = 0.0f;
		private float myMouthValue = 0.0f;
		private float myMouthVelocity = 0.0f;
		private float myLungScale = 0.0f;
		private float myLungValuePrev = 0.0f;
		private bool myLungWasDown = true;
		private float myNoseScale = 0.0f;
		private float myNoseSign = 0.0f;
		private float myMouthScale = 0.0f;
		private bool myMouthWasClosing = true;

		private BlendValue myLungValue;
		private BlendValue myBreathInValue;
		private BlendValue myBreathOutValue;
		private BlendValue myBreathMouthIOValue;
		private BlendValue myBreathMouthOIValue;
		private BlendValue myDebugValue1;
		private BlendValue myDebugValue2;
		private BlendValue myDebugValue3;

		public override void Init()
		{
			myBreathing = FindWithinSamePlugin<Breathing>(this);
			InitUI();
			InitAnimation();
			InitEvents();
		}

		private void InitUI()
		{
			Utils.OnInitUI(CreateUIElement);
			Utils.SetupInfoTextNoScroll(this,
				"<color=#606060><size=40><b>DriverBreathing</b></size>\nThis animation driver controls chest, stomach, nose and mouth animation based on breathing.</color>\n\n" +
				"<b>ChestMorph Min/Max:</b>\nMinimum/Maximum values for chest morph.\n\n" +
				"<b>ChestJointDrive Min/Max:</b>\nMinimum/Maximum values for 'Joint Drive X Angle' of chest control.\n\n" +
				"<b>ChestJointDrive Spring:</b>\nValue for 'Joint Drive Spring' of chest control.\n\n" +
				"<b>Stomach Min/Max:</b>\nMinimum/Maximum values for stomach morph.\n\n" +
				"<b>MouthMorph Min/Max:</b>\nMinimum/Maximum values for MouthOpen morph.\n\n" +
				"<b>Mouth Open/Close Time:</b>\nTime it takes for the mouth to open/close.\n\n" +
				"<b>LipsMorph Max:</b>\nMaximum value for lips morph when the mouth is fully open.\n\n" +
				"<b>NoseInMorph Max:</b>\nMaximum value for nose morph when breathing in.\n\n" +
				"<b>NoseOutMorph Max:</b>\nMaximum value for nose morph when breathing out.\n\n",
				1350.0f, true
			);

			myChestMorphMin = SetupSliderFloat(this, "ChestMorph Min", -0.1f, -1.0f, 1.0f, false);
			myChestMorphMax = SetupSliderFloat(this, "ChestMorph Max",  0.5f, -1.0f, 1.0f, false);
			myChestJointMin = SetupSliderFloat(this, "ChestJointDrive Min", -2.0f, -20.0f, 20.0f, false);
			myChestJointMax = SetupSliderFloat(this, "ChestJointDrive Max", 3.0f, -20.0f, 20.0f, false);
			myChestSpring = SetupSliderFloat(this, "ChestJointDrive Spring", 160.0f, 0.0f, 250.0f, false);
			myStomachMin = SetupSliderFloat(this, "StomachMin", -0.1f, -1.0f, 1.0f, false);
			myStomachMax = SetupSliderFloat(this, "StomachMax",  0.4f, -1.0f, 1.0f, false);
			myMouthOpenMin = SetupSliderFloat(this, "MouthMorph Min", 0.0f, 0.0f, 1.0f, false);
			myMouthOpenMax = SetupSliderFloat(this, "MouthMorph Max", 0.6f, 0.0f, 1.0f, false);
			myMouthOpenTime = SetupSliderFloat(this, "Mouth Open Time", 0.05f, 0.0f, 1.0f, false);
			myMouthCloseTime = SetupSliderFloat(this, "Mouth Close Time", 0.2f, 0.0f, 1.0f, false);
			myLipsMax = SetupSliderFloat(this, "LipsMorph Max", 1.0f, -1.0f, 1.0f, false);
			myNoseInMax = SetupSliderFloat(this, "NoseInMorph Max", 0.6f, 0.0f, 1.0f, false);
			myNoseOutMax = SetupSliderFloat(this, "NoseOutMorph Max", 0.25f, 0.0f, 1.0f, false);

			if (!Utils.IsInPackage(this))
			{
				myDebug1 = SetupSliderFloat(this, "Debug1", 0.0f, 0.0f, 1.0f, true);
				myDebug2 = SetupSliderFloat(this, "Debug2", 0.0f, 0.0f, 1.0f, true);
				myDebug3 = SetupSliderFloat(this, "Debug3", 0.0f, 0.0f, 1.0f, true);
			}
        }
		
		private string GetPackagePath()
		{		
            string id = name.Substring(0, name.IndexOf('_'));
            string filename = manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}

		private void InitAnimation()
		{
			myGeometry = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;
			string morphDirMale = "Custom/Atom/Person/Morphs/male/MacGruber/Pose/Breathing/";
			string morphDirFemale = "Custom/Atom/Person/Morphs/female/MacGruber/Pose/Breathing/";
			ResetMorphs(myGeometry?.morphsControlMaleUI, morphDirMale);
			ResetMorphs(myGeometry?.morphsControlFemaleUI, morphDirFemale);

			myGender = myGeometry != null ? myGeometry.gender : DAZCharacterSelector.Gender.None;
			GenerateDAZMorphsControlUI morphControl = myGeometry?.morphsControlUI;
			if (morphControl != null)
			{
				string dirGender = (myGender == DAZCharacterSelector.Gender.Male ? morphDirMale : morphDirFemale);
				string morphDirOriginal = Utils.GetPackagePath(this) + dirGender;
				string morphDirCustom = GetPackagePath() + dirGender;
				
				myChestMorph   = morphControl.GetMorphByUid(morphDirOriginal + "Breathing_Chest.vmi");
				myStomachMorph = morphControl.GetMorphByUid(morphDirOriginal + "Breathing_Stomach.vmi");
				myLipsMorph    = morphControl.GetMorphByUid(morphDirOriginal + "Breathing_Lips.vmi");
				myNoseInMorph  = morphControl.GetMorphByUid(morphDirOriginal + "Breathing_NoseIn.vmi");
				myNoseOutMorph = morphControl.GetMorphByUid(morphDirOriginal + "Breathing_NoseOut.vmi");
				myMouthMorph   = morphControl.GetMorphByUid(morphDirCustom + "AA_Mouth_Open_Driver.vmi");
			}
			myChestControl = containingAtom.GetStorableByID("chestControl") as FreeControllerV3;

			myAnimationValid =
				myChestMorph     != null &&
				myStomachMorph   != null &&
				myMouthMorph     != null &&
				myLipsMorph      != null &&
				myNoseInMorph    != null &&
				myNoseOutMorph   != null &&
				myChestControl   != null;
		}

		private void ResetMorphs(GenerateDAZMorphsControlUI morphControl, string morphDir)
		{
			if (morphControl == null)
				return;

			List<DAZMorph> morphs = morphControl.GetMorphs();
			for (int i=0; i<morphs.Count; ++i)
			{
				if (morphs[i].uid.Contains(morphDir))
					morphs[i].Reset();
			}
		}

		private void OnDisable()
		{
			if (myAnimationValid)
			{
				myChestControl.jointRotationDriveXTarget = 0.0f;
				myChestMorph.Reset();
				myStomachMorph.Reset();
				myMouthMorph.Reset();
				myLipsMorph.Reset();
				myNoseInMorph.Reset();
				myNoseOutMorph.Reset();
			}
		}

		private void OnDestroy()
        {
			Utils.OnDestroyUI();
        }

		private void InitEvents()
		{
			{
				BlendEvent value1 = myBreathing.RegisterBlendEvent(
					SM(MRK_BreathIn),
					SM(MRK_HoldIn),
					SM(MRK_BreathEnd)
				);
				BlendEvent value2 = myBreathing.RegisterBlendEvent(
					SM(MRK_BreathOut),
					SM(MRK_HoldOut),
					SM(MRK_BreathEnd)
				);

				myLungValue = new BlendCombineLinearInOut(value1, value2);

				myDebugValue1 = value1;
				myDebugValue2 = value2;
				myDebugValue3 = myLungValue;
			}

			{
				BlendEvent value1 = myBreathing.RegisterBlendEvent(
					SM(MRK_BreathIn, -0.12f),
					RM(MRK_BreathIn, MRK_HoldIn, 0.70f, 0.00f, 1.00f),
					SM(MRK_BreathEnd)
				);
				BlendEvent value2 = myBreathing.RegisterBlendEvent(
					RM(MRK_HoldIn, MRK_BreathIn, 0.30f, -0.40f, -0.05f),
					SM(MRK_HoldIn,  0.1f),
					SM(MRK_BreathEnd)
				);
				myBreathInValue = new BlendCombineLinearInOut(value1, value2);
			}

			{
				BlendEvent value1 = myBreathing.RegisterBlendEvent(
					SM(MRK_BreathOut, -0.12f),
					RM(MRK_BreathOut, MRK_HoldOut, 0.30f, 0.0f, 0.2f),
					SM(MRK_BreathEnd)
				);
				BlendEvent value2 = myBreathing.RegisterBlendEvent(
					RM(MRK_HoldOut, MRK_BreathOut, 0.70f, -2.0f, 0.0f),
					SM(MRK_HoldOut,  0.05f),
					SM(MRK_BreathEnd)
				);
				myBreathOutValue = new BlendCombineLinearInOut(value1, value2);
			}

			{
				BlendEvent value1 = myBreathing.RegisterBlendEvent(
					SM(MRK_BreathIn, 0.0f),
					RM(MRK_BreathIn, MRK_HoldIn, 0.50f, 0.0f, 0.3f),
					SM(MRK_BreathEnd)
				);
				BlendEvent value2 = myBreathing.RegisterBlendEvent(
					RM(MRK_HoldOut, MRK_BreathOut, 0.70f, -1.0f, 0.0f),
					SM(MRK_HoldOut,  0.0f),
					SM(MRK_BreathEnd)
				);
				myBreathMouthIOValue = new BlendCombineLinearInOut(value1, value2);
			}

			{
				BlendEvent value1 = myBreathing.RegisterBlendEvent(
					SM(MRK_BreathOut, 0.0f),
					RM(MRK_BreathOut, MRK_HoldOut, 0.30f, 0.0f, 0.1f),
					SM(MRK_BreathEnd)
				);
				BlendEvent value2 = myBreathing.RegisterBlendEvent(
					RM(MRK_HoldIn, MRK_BreathIn, 0.30f, -0.1f, 0.0f),
					SM(MRK_HoldIn, 0.0f),
					SM(MRK_BreathEnd)
				);
				myBreathMouthOIValue = new BlendCombineLinearInOut(value1, value2);
			}
		}

		private void Update()
		{
			if (myGeometry == null)
				return;
			if (myGeometry.gender != myGender)
				InitAnimation();

			if (myDebug1 != null && myDebugValue1 != null)
				myDebug1.val = myDebugValue1.Value;
			if (myDebug2 != null && myDebugValue2 != null)
				myDebug2.val = myDebugValue2.Value;
			if (myDebug3 != null && myDebugValue3 != null)
				myDebug3.val = myDebugValue3.Value;

			Breath breath = myBreathing.Get(CURR);
			if (!breath.IsValid() || !myAnimationValid)
				return;

			{
				float lungValue = myLungValue.Value;
				if (lungValue >= myLungValuePrev && myLungWasDown)
					myLungScale = breath.Depth * breath.Power * breath.Lung;
				myLungWasDown = lungValue < myLungValuePrev;
				myLungValuePrev = lungValue;
				lungValue = Mathf.SmoothStep(0.0f, myLungScale, lungValue);

				// correct a bit for the chest and stomach morphs being essentially linear scale, moving faster near max range
				float lungValueCorrected = Mathf.Pow(lungValue, 0.9f);

				// Chest morphs
				myChestMorph.morphValue = Mathf.Lerp(myChestMorphMin.val, myChestMorphMax.val, lungValueCorrected);

				// Chest joint rotation
				float chestJoint = Mathf.Lerp(myChestJointMin.val, myChestJointMax.val, lungValue);
                myChestControl.jointRotationDriveXTarget = chestJoint;
				myChestControl.jointRotationDriveDamper = 1.0f;
				myChestControl.jointRotationDriveSpring = myChestSpring.val;

				// Stomach morph
				myStomachMorph.morphValue = Mathf.Lerp(myStomachMin.val, myStomachMax.val, lungValueCorrected);
            }

			{
				float mouth = 0.0f;
				float nose = 0.0f;
				bool hasNoseIn = breath.Entry.HasTag(TAG_NOSEIN);
				bool hasNoseOut = breath.Entry.HasTag(TAG_NOSEOUT);
				if (hasNoseIn && hasNoseOut)
				{
					nose = myBreathInValue.Value - myBreathOutValue.Value;
				}
				else if (hasNoseIn)
				{
					nose = myBreathInValue.Value;
					mouth = myBreathOutValue.Value;
				}
				else if (hasNoseOut)
				{
					nose = -myBreathOutValue.Value;
					mouth = myBreathInValue.Value;
				}
				else
				{
					mouth = (breath.Entry.Format == FORMAT_INOUT) ? myBreathMouthIOValue.Value : myBreathMouthOIValue.Value;
				}

				// Nose morphs
				float noseSign = Mathf.Sign(nose);
				if (myNoseSign != noseSign)
				{
					myNoseScale = breath.Depth * breath.Power * breath.Nose;
					myNoseSign = noseSign;
				}
				if (nose >= 0.0f)
				{
					myNoseInMorph.morphValue = myNoseScale*myNoseInMax.val*nose;
					myNoseOutMorph.morphValue = 0.0f;
				}
				else
				{
					myNoseInMorph.morphValue = 0.0f;
					myNoseOutMorph.morphValue = -myNoseScale*myNoseOutMax.val*nose;
				}

				// Mouth & Lips morphs
				float direction = (hasNoseIn || hasNoseOut) ? -1.0f : 1.0f;
				myMouthOpen = Mathf.Clamp01(myMouthOpen + direction * Time.deltaTime);
				mouth = Mathf.Lerp(mouth, mouth*0.8f+0.2f, myMouthOpen);
				float mouthSmoothTime = mouth < myMouthValue ? myMouthCloseTime.val : myMouthOpenTime.val;
				mouth = Mathf.SmoothDamp(myMouthValue, mouth, ref myMouthVelocity, mouthSmoothTime);
				if (mouth >= myMouthValue && myMouthWasClosing)
					myMouthScale = breath.Depth * breath.Power * breath.Mouth;
				myMouthWasClosing = mouth < myMouthValue;
				myMouthValue = mouth;
				myMouthMorph.morphValue = Mathf.SmoothStep(myMouthOpenMin.val, myMouthScale * myMouthOpenMax.val, mouth);
				myLipsMorph.morphValue = Mathf.SmoothStep(0.0f, myLipsMax.val, mouth);
			}
		}
	}
}